//package Disease;

//import Person.*;
//import BayesNetwork.*;
//import GenExceptions.*;

import java.util.*;

public class AutosomalDisease extends Disease {

private String name;
private String[] alleles;
private String[] genotypes;
private String[] phenotypes;
private double[] gDefaultProbs;

private boolean bDomRec;

//constructor
	public AutosomalDisease(String name, String[] Alleles, String[] Phenotypes, double[] probs) 
		throws InvalidDiseaseException {
		// requires:  None of the inputs can be null.  Phenotypes[0] is the phenotype for
		//      the genotype Allele1-Allele1, Phenotype[1] is for Allele1-Allele2, and Phenotype[2]
		//		is for Allele2-Allele2.
		// @return  Creates an autosomal disease with the two alleles in Alleles, which
		//		have the 3 phenotypes in Phenotypes where Phenotypes[0] is the phenotype for
		//      the genotype Allele1-Allele1, Phenotype[1] is for Allele1-Allele2, and Phenotype[2]
		//		is for Allele2-Allele2.  probs are the probabilities for of each phenotypes (and its
		//		genotypes since 100% penetrance).
		// @exception	An InvalidDiseaseException is thrown if alleles does not contain 2 different alleles,
		//		if phenotypes does not contain 3 items, if Phenotype[0] = phenotype[2], 
		//		if probs does not contain 3 doubles that sum close to 1.

		initAutosomalDisease(name, Alleles, Phenotypes, probs);
	}

	private void initAutosomalDisease(String name, String[] Alleles, String[] Phenotypes, double[] probs) 
		throws InvalidDiseaseException {

		this.name = name;

		if ((probs.length != 3) || (Alleles.length != 2) || (Alleles[0].equals(Alleles[1])))
			throw new InvalidDiseaseException("The number of components given to create the disease is incorrect");

		// figure out if this is domrec - if so put dom allele as first
		if ((Phenotypes.length != 3) || (Phenotypes[0].equals(Phenotypes[2])))
			throw new InvalidDiseaseException();
		if (Phenotypes[0].equals(Phenotypes[1])) {
			// this is domrec- first allele is dominant
			bDomRec = true;
			this.alleles = (String[])Alleles.clone();
			this.phenotypes = new String[2];
			this.phenotypes[0] = Phenotypes[0];
			this.phenotypes[1] = Phenotypes[2];
			this.gDefaultProbs = (double[])probs.clone();
		}
		else if (Phenotypes[1].equals(Phenotypes[2])) {
			bDomRec = true;
			this.alleles = new String[2];
			this.alleles[0] = Alleles[1];
			this.alleles[1] = Alleles[0];
			this.phenotypes = new String[2];
			this.phenotypes[0] = Phenotypes[2];
			this.phenotypes[1] = Phenotypes[0];
			this.gDefaultProbs = new double[3];
			this.gDefaultProbs[0] = probs[2];
			this.gDefaultProbs[1] = probs[1];
			this.gDefaultProbs[2] = probs[0];
		}
		else {
			// this is a codom disease
			bDomRec = false;
			this.alleles = (String[])Alleles.clone();
			this.phenotypes = (String[])Phenotypes.clone();
			this.gDefaultProbs = (double[])probs.clone();
		}

		genotypes = new String[3];
		genotypes[0] = this.alleles[0] + "-" + this.alleles[0];
		genotypes[1] = this.alleles[0] + "-" + this.alleles[1];
		genotypes[2] = this.alleles[1] + "-" + this.alleles[1];

		

		// make sure probs are valid 
		if (!super.sumToOne(probs))
			throw new InvalidDiseaseException();

		sA1A1_A1A1 = new String[2];
		sA1A1_A1A1[0] = genotypes[0];
		sA1A1_A1A1[1] = genotypes[0];
		sA1A1_A1A2 = new String[2];
		sA1A1_A1A2[0] = genotypes[0];
		sA1A1_A1A2[1] = genotypes[1];
		sA1A1_A2A2 = new String[2];
		sA1A1_A2A2[0] = genotypes[0];
		sA1A1_A2A2[1] = genotypes[2];
		sA1A2_A1A1 = new String[2];
		sA1A2_A1A1[0] = genotypes[1];
		sA1A2_A1A1[1] = genotypes[0];
		sA1A2_A1A2 = new String[2];
		sA1A2_A1A2[0] = genotypes[1];
		sA1A2_A1A2[1] = genotypes[1];
		sA1A2_A2A2 = new String[2];
		sA1A2_A2A2[0] = genotypes[1];
		sA1A2_A2A2[1] = genotypes[2];
		sA2A2_A1A1 = new String[2];
		sA2A2_A1A1[0] = genotypes[2];
		sA2A2_A1A1[1] = genotypes[0];
		sA2A2_A1A2 = new String[2];
		sA2A2_A1A2[0] = genotypes[2];
		sA2A2_A1A2[1] = genotypes[1];
		sA2A2_A2A2 = new String[2];
		sA2A2_A2A2[0] = genotypes[2];
		sA2A2_A2A2[1] = genotypes[2];
	}

	public AutosomalDisease(String sDisease) throws InvalidDiseaseException{
	  /** requires: sDisease is not null.
	   * @return an autosomal disease based on the specification in the string
	   * @exception An invalidDiseaseException is thrown if the string contains an invalid autosomalDisease.
	   * A valid sDisease has the format:
	   *"AutosomalDisease" String <name> <Allele 1> <Allele2> <Phenotype 1> <Phenotype 2> <Phenotype 3> double <Prob 1> <Prob 2> <Prob 3>
	   **/
		StringTokenizer st = new StringTokenizer(sDisease);
		try {
			String sNext = st.nextToken();
			if (!sNext.equals("AutosomalDisease"))
				throw new InvalidDiseaseException("Should be an Autosomal Disease");

			String name = st.nextToken();

			String[] sAlleles = new String[2];
			sAlleles[0] = st.nextToken();
			sAlleles[1] = st.nextToken();
			String[] sPhenotypes = new String[3];
			sPhenotypes[0] = st.nextToken();
			sPhenotypes[1] = st.nextToken();
			sPhenotypes[2] = st.nextToken();

			double[] probs = new double[3];
			probs[0] = Double.valueOf(st.nextToken()).doubleValue();
			probs[1] = Double.valueOf(st.nextToken()).doubleValue();
			probs[2] = Double.valueOf(st.nextToken()).doubleValue();

			initAutosomalDisease(name, sAlleles, sPhenotypes, probs);
		}
		catch(NoSuchElementException e) {
			throw new InvalidDiseaseException("Not enough components in the string to create an autosomal disease");
		}
		catch(NumberFormatException e) {
			throw new InvalidDiseaseException("The probabilities provided are not numbers");
		}

	}

	public AutosomalDisease() {
		// effects:  Creates an autosomal disease with domiant allele = A, recessive allele = a,
		//			 dominant phenotype = "affected", recessive phenotype = "unaffected" and
		//			 genotype probs of 0.25, 0.5, and 0.25 for genotypes AA, Aa, and aa respectively
		String[] sAlleles = new String[2];
		sAlleles[0] = "A";
		sAlleles[1] = "a";
		String[] sPhenotypes = new String[3];
		sPhenotypes[0] = sPhenotypes[1] = "Affected";
		sPhenotypes[2] = "Unaffected";
		double[] probs = new double[3];
		probs[0] = .25;
		probs[1] = .5;
		probs[2] = .25;
		try {
			initAutosomalDisease("Default Autosomal Disease", sAlleles, sPhenotypes, probs);
		}
		catch(InvalidDiseaseException e) {
			System.err.println("Unexpected InvalidDiseaseException in AutosomalDisease()");
		}
	}

// methods

  /** @return The name of this disease.
   **/
  public String getName() {
		return name;
	}

  /** @return The possible genotypes of a person of gender sexType.
   **/
	public String[] getGenotypes(int sexType) {
		
		return (String[])genotypes.clone();
	}

  /** @return The possible phenotypes of a person of gender sexType.
   **/
	public String[] getPhenotypes(int sexType) {
		return (String[])phenotypes.clone();
	}

  public String getDefaultPhenotype(int sexType) {
		return phenotypes[0];
	}

	//********* here are some local variables used in methods to manipulate Bayes Net

	private final double[] gA1A1_A1A1probs = {1, 0, 0};
	private final double[] gA1A1_A1A2probs = {0.5, 0.5, 0};
	private final double[] gA1A1_A2A2probs = {0, 1, 0};
	private final double[] gA1A2_A1A2probs = {.25, .5, .25};
	private final double[] gA1A2_A2A2probs = {0, .5, .5};
	private final double[] gA2A2_A2A2probs = {0, 0, 1};

	private String[] sA1A1_A1A1;
	private String[] sA1A1_A1A2;		
	private String[] sA1A1_A2A2;		
	private String[] sA1A2_A1A1;		
	private String[] sA1A2_A1A2;		
	private String[] sA1A2_A2A2;		
	private String[] sA2A2_A1A1;		
	private String[] sA2A2_A1A2;		
	private String[] sA2A2_A2A2;
		

	//private final double[] phAa = {1, 0};
	//private final double[] phaa = {0, 1};

  /** @requires bn and indiv are not null.
      @return indiv is added to bn and indiv's pehnotype and genotype nodes are created.
  **/
	public void addPerson(BayesNetwork bn, Person indiv) {
		// create the two nodes
		try {
			BayesNetworkNode genotypeNode = bn.createNode(getGenotypes(indiv.getSex()), gDefaultProbs);
			indiv.setGenotypeNode(genotypeNode);
			BayesNetworkNode phenotypeNode;
			if (bDomRec) {
				double[] tempPhenProbs = {1, 0};  // this is temporary
				phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
			}
			else {
				double[] tempPhenProbs = {1, 0, 0};  // this is temporary
				phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
			}
			indiv.setPhenotypeNode(phenotypeNode);
			bn.createArc(genotypeNode, phenotypeNode);
			BayesNetworkNode[] bnns = {genotypeNode};
			String[] sphA1A1 = {genotypes[0]};
			String[] sphA1A2 = {genotypes[1]};
			String[] sphA2A2 = {genotypes[2]};

			if (bDomRec) {
				double[] phA1A1 = {1, 0};
				double[] phA1A2 = {1, 0};
				double[] phA2A2 = {0, 1};
				phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
				phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
				phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
			}
			else  {
				double[] phA1A1 = {1, 0, 0};
				double[] phA1A2 = {0, 1, 0};
				double[] phA2A2 = {0, 0, 1};
				phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
				phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
				phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
			}

			phenotypeNode.makeObservation(getDefaultPhenotype(indiv.getSex()));
		}
		catch(ProbabilityException e) {
			System.err.println("Disease add person had a prob exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add person had a value exception");
		}
		catch(InvalidArcException e) {
			System.err.println("Disease add person had a arc exception");
		}
	}

  /** @requires bn and indiv are not null.
      @return indiv is removed from bn and invid's phenotype and genotype nodes are set to null.
      If indiv is connected, nothing is done. 
   **/
	public void removePerson(BayesNetwork bn, Person indiv) {
		// make sure there are no children or parents of genotype
		BayesNetworkNode genotypeNode = indiv.getGenotypeNode();
		Enumeration enumParents = genotypeNode.getParents();
		if (enumParents.hasMoreElements())
			return;  // don't do anything if has indiv has parents
		BayesNetworkNode phenotypeNode = indiv.getPhenotypeNode();
		Enumeration enumChildren = genotypeNode.getChildren();
		// must have at least one child, phenotype node
		enumChildren.nextElement();
		if (enumChildren.hasMoreElements())
			return;  // indiv has children

		bn.removeNode(phenotypeNode);
		bn.removeNode(genotypeNode);

		indiv.setPhenotypeNode(null);
		indiv.setGenotypeNode(null);
	}
  /** @requires bn, child, parent1, parent2 are not null.
      All the people have been added to the bn and have valid genotype and phenotype nodes.
      This child is valid (parents are of opposite sex) and loops are created by the addition.
      @return Arcs and probabilistic relationare estabilished in bn to represent the child relationship.
  **/
	public void addChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// try making links
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();
		try {
			bn.createArc(parentNode1, childNode);
			bn.createArc(parentNode2, childNode);
		}
		catch (InvalidArcException e) {
			System.err.println("Disease.addChild had an arc exception");
			return;  // created a loop - time to bail
		}

		// need to reinit child genotypenode
		try {
			BayesNetworkNode[] bnns = {parentNode1, parentNode2};
			childNode.changeProbFunction(bnns, sA1A1_A1A1, gA1A1_A1A1probs);
			childNode.changeProbFunction(bnns, sA1A1_A1A2, gA1A1_A1A2probs);
			childNode.changeProbFunction(bnns, sA1A1_A2A2, gA1A1_A2A2probs);
			childNode.changeProbFunction(bnns, sA1A2_A1A1, gA1A1_A1A2probs);
			childNode.changeProbFunction(bnns, sA1A2_A1A2, gA1A2_A1A2probs);
			childNode.changeProbFunction(bnns, sA1A2_A2A2, gA1A2_A2A2probs);
			childNode.changeProbFunction(bnns, sA2A2_A1A1, gA1A1_A2A2probs);
			childNode.changeProbFunction(bnns, sA2A2_A1A2, gA1A2_A2A2probs);
			childNode.changeProbFunction(bnns, sA2A2_A2A2, gA2A2_A2A2probs);
		}
		catch(ProbabilityException e) {
			System.err.println("Disease add child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add child had a value exception.");
		}
	}

  /** @requires bn, child, parent1 and parent2 are not null.
      All the people have been added to the bn and have valid genotype and phenotype nodes.
      @return Arcs and probabilistic relations representing the child are removed in bn.
  **/
	public void removeChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// remove the links if they exist
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();

		bn.removeArc(parentNode1, childNode);
		bn.removeArc(parentNode2, childNode);

		// init the genotype node again
		try {
			childNode.changeProbFunction(null, null, gDefaultProbs);
		}
		catch(ProbabilityException e) {
			System.err.println("Disease remove child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease remove child had a value exception");
		}
	}

  /** @return A string representation of this disease, which could be passed back into the createDisease function to produce an exact clone.
   **/
	public String toString() {
		String s = "AutosomalDisease\n";
		s += name + "\n";
		s += alleles[0] + "\n";
		s += alleles[1] + "\n";
		s += phenotypes[0] + "\n";
		if (bDomRec) {
			s += phenotypes[0] + "\n";
			s += phenotypes[1] + "\n";
		}
		else {
			s += phenotypes[1] + "\n";
			s += phenotypes[2] + "\n";
		}
		for (int x = 0; x < gDefaultProbs.length; x++)
			s += gDefaultProbs[x] + "\n";

		return s;
	}

	// **** methods specific to an autosomal disease

	public String[] getAlleles() {
		return (String[])alleles.clone();
	}

	public String[] getPhenotypes() {
		return (String[])phenotypes.clone();
	}

  public double[] getGenoProbs() {
		return (double[])gDefaultProbs.clone();
	}

  public double[][] getExpressionRates(int sexType){
    // requires: 100% penetrance
    // effects: Returns the expression rates from the possible genotypes
    
    double[][] expressionRates = new double[1][3];
    if (bDomRec) { // if domrec
      expressionRates[0][0] = 1; // AA
      expressionRates[0][1] = 1; // Aa
      expressionRates[0][2] = 0; // aa
    }
    else{ // if codom
      expressionRates[0][0] = 1; // AA
      expressionRates[0][1] = 0.5; // Aa
      expressionRates[0][2] = 0; // aa
    }
    return expressionRates;
  }

  public boolean isAgeDependent(){
    //effects: Returns false for now
    return false;
  }

  public double[][] getFrequencyRates(){
    //effects: Returns 0.1 for all alleles for now.  Should return the frequency of each allele in the order from getDefectiveAlleles()
    double[][] frequencyRates = new double[1][2];
    frequencyRates[0][0] = 0.1;
    frequencyRates[0][1] = 0.1;
    return frequencyRates;
  }

  public String[] getRaces(){
    //effects: returns null for now.
    return null;
  }

  public boolean isRaceDependent(){
    //effects: returns false for now.
    return false;
  }

  public String getNormalAllele(){
    //effects: returns dominant allele for now.
    if (bDomRec){ //domrec
      return alleles[0];
    }
    else{ //codom
      return alleles[0];
    }
  }

  public String[] getDefectiveAlleles(){
    //effects: returns all alleles for now.
    return (String[])alleles.clone();
  }

  public boolean isSexLinked(){
    //effects: returns false for now.
    return false;
  }

  public boolean hasMutationRate(){
    //effects: returns false for now.
    return false;
  }

  public double[] getMutationRates(){
    //effects: returns 0 mutation rate for one defective allele
    double[] mutationRates = new double[1];
    mutationRates[0] = 0;
    return mutationRates;
  }
  
}
