//package Disease;

//import Person.*;
//import BayesNetwork.*;

import java.util.*;

abstract public class Disease {

	abstract public String getName();
	// effects: Return the name of this disease

	abstract public String[] getGenotypes(int pPersonSexType);
	// effects:   Returns the posible genotypes of a person of type pPersonSexType

	abstract public String[] getPhenotypes(int pPersonSexType);
	// effects:   Returns the posible phenotypes of a person of type pPersonSexType

	abstract public void addPerson(BayesNetwork bn, Person indiv);
	// requires:  bn and indiv are not null
	// effects:   This individual is added to bn, and the indiv's phenotype and genotype nodes are created.

	abstract public void removePerson(BayesNetwork bn, Person indiv);
	// requires:  bn and indiv are not null
	// effects:   This individual is removed from this bn, and the indiv's phenotype and genotype nodes 
	//		are set to null.  It person has children or parents, nothing is done.

	abstract public void addChild(BayesNetwork bn, Person child, Person parent1, Person parent2);
	// requires:  bn and the people are not null.  All the people must have been added to the bn and have
	//			valid genotype and phenotype nodes.  Requires that this child is valid (parents are oppsite
	//			sex) and no loops are created.
	// effects:   Arcs and probabilistic relations are established in bn to represent the child
	//  relationship.

	abstract public void removeChild(BayesNetwork bn, Person child, Person parent1, Person parent2);
	// requires:  bn and the people are not null.  All the people must have been added to the bn and have
	//			valid genotype and phenotype nodes.
	// effects:   Arcs and probabilistic relations are removed in bn to for the child is it was a child of these
	//    parents.

	abstract public String toString();
	// effects:   A string representation of this disease, which could be passed back to the Disease.createDisease
	//  to produce an exact clone.

	abstract public String getNormalAllele();
	// effects:  Returns the normal allele for this disease

	abstract public String[] getDefectiveAlleles();
	// effects:  Returns the defective alleles of this disease

	abstract public boolean isSexLinked();
	// effects:  Returns true if the disease is xlinked

	abstract public boolean hasMutationRate();
	// effects:	 Returns true if this disease has a mutation rate

	abstract public double[] getMutationRates();
	// effects:  Returns the rate of mutation of the normal allele to each bad allele, where each entry
	//			 in the array corresponds to the bad allele in the array returned by getBadAlleles()

	abstract public boolean isRaceDependent();
	// effects:  Returns true if this diseases frequencies depend on race

	abstract public String[] getRaces();
	// effects:  Returns the races of this disease if race dependent, otherwise null

	abstract public double[][] getFrequencyRates();
	// effects:  Returns the frequency of each allele, where the first entry is normal allele, and each other
	//			 entry is for the defective alleles in order from getDefectiveAlleles()
	//			 Each row is for the races, in the order of getRaces.
	//			 If this is not race dependent, there is only one row

	abstract public boolean isAgeDependent();
	// effects:  Returns true if this diseases expression depends on age
	
	abstract public double[][] getExpressionRates(int sexType);
	// effects:  Returns the expression rates from the possible genotypes to the affected condition for a given sex.
	//		     Each row represents an age for the age groups starting with 0-9.
	//			 Each entry in a row corresponds to the genotype returned by getGenotypes(sex)
	//			 There is only one row for disease not age dependent
	static public Disease createDisease(String sDisease) throws InvalidDiseaseException {
		// requires:  sDisease is not null, the type of the disease is given as the first token of sDisease
		// effects:   Creates a disease with the specification in sDisease.  There is support
		//			autosomal diseases.  An InvalidDiseaseException is thrown
		//			If no valid disease is described in sDisease.
	  StringTokenizer st = new StringTokenizer(sDisease, "\n");
	  String firstToken = st.nextToken();
	  if (firstToken.equals("NewAutosomalDisease")){
		NewAutosomalDisease AD = new NewAutosomalDisease(sDisease);
			return AD;
	  }
	  else if (firstToken.equals("XlinkedDisease")){
	    XlinkedDisease XLD = new XlinkedDisease(sDisease);
	    return XLD;
	  }
	  else if (firstToken.equals("NewXlinkedDisease")){
	    NewXlinkedDisease NXLD = new NewXlinkedDisease(sDisease);
	    return NXLD;
	  }
	  else{
	    throw new InvalidDiseaseException("cannot determine type of disease");
	  }
	}

	protected boolean sumToOne(double[] dArray) {
		// requires:  dArray is not null
		// effects:  return true if dArray elements sum to one and are are non-negative
		double dSum = 0;
		for (int x=0; x < dArray.length; x++) {
			if (dArray[x] < 0)
				return false;
			else
				dSum += dArray[x];
		}
		if (java.lang.Math.abs(dSum - 1.0) > .1)
			return false;
		else 
			return true;
	}

}
