//package Disease;

//import Person.*;
//import BayesNetwork.*;
//import GenExceptions.*;

import java.util.*;

public class NewAutosomalDisease extends Disease {
	final int NUM_AGE_GROUPS = 9;

	String name;
	String[] badAlleles;
	String normalAllele;
	double[] mutationRates;
	double[][] expressionTable;
	Vector vRaces;
	double[][] alleleFreqRates;

	// these are locals created by init
	String[] genotypes;
	double[][] genotypeFreqs;  // this is the race/ genotype frequency calculated from alleleFreq
	double[][] inheritenceTable;
	double noMutationRate;  // this is the rate by which the normal allele does not mutatate
	boolean bHasMutationRate;
	boolean bAgeDependentExpression;
	boolean bRaceDependent;
	Hashtable mutationAlleleIndexTable;

	public NewAutosomalDisease(String name, String normalAllele,  String[] badAlleles,
		double[] mutationRates, double[][] expressionRates, Vector vRaces, double[][] freqRates) {

		this.name = name;
		this.normalAllele = normalAllele;
		this.badAlleles = badAlleles;

		this.mutationRates = mutationRates;
		this.expressionTable = expressionRates;
		this.vRaces = vRaces;
		this.alleleFreqRates = freqRates;
		// DEBUG
		System.out.println("Creating New Autosomal Disease");
		init();
	}

	public NewAutosomalDisease(String s) throws InvalidDiseaseException {
		StringTokenizer st = new StringTokenizer(s, "\n");

		try {
			String sNext = st.nextToken();
			if (!sNext.equals("NewAutosomalDisease"))
				throw new InvalidDiseaseException();

			this.name = st.nextToken();

			this.normalAllele = st.nextToken();

			int iNumBadAlleles = Integer.valueOf(st.nextToken()).intValue();
			if (iNumBadAlleles < 1)
				throw new InvalidDiseaseException();

			this.badAlleles = new String[iNumBadAlleles];
			for (int x= 0; x < iNumBadAlleles; x++)
				this.badAlleles[x] = st.nextToken();

			sNext = st.nextToken();
			if (sNext.equals("Mutation Rates")) {
				// the number of mutations is equals to number of bad alleles
				this.mutationRates = new double[iNumBadAlleles];
				for (int x = 0; x < iNumBadAlleles; x++)
					this.mutationRates[x] = Double.valueOf(st.nextToken()).doubleValue();

				// get the next string for age dependency
				sNext = st.nextToken();
			} // else no mutation rates

			int iNumOfAlleles = iNumBadAlleles + 1;
			int iNumGenotypes = iNumOfAlleles * (iNumOfAlleles + 1)/2;
			if (sNext.equals("Age dependent")) {
				this.expressionTable = new double[NUM_AGE_GROUPS][iNumGenotypes];
				for (int x= 0; x < this.expressionTable.length; x++) {
					for (int y = 0; y < this.expressionTable[x].length; y++)
						this.expressionTable[x][y] = Double.valueOf(st.nextToken()).doubleValue();
				}
			}
			else if (sNext.equals("Not age dependent")) {
				this.expressionTable = new double[1][iNumGenotypes];
				for (int x=0; x < this.expressionTable[0].length; x++)
					this.expressionTable[0][x] = Double.valueOf(st.nextToken()).doubleValue();
			}
			else
				throw new InvalidDiseaseException();

			sNext = st.nextToken();
			if (sNext.equals("Race dependent")) {
				int iNumRaces = Integer.valueOf(st.nextToken()).intValue();
				this.vRaces = new Vector();
				for (int x=0; x < iNumRaces; x++)
					this.vRaces.addElement(st.nextToken());

				// next string for freq
				sNext = st.nextToken();
			}
		
			if (!sNext.equals("Freq rates"))
				throw new InvalidDiseaseException();

			// the format for freqRates is one table entry per race, with the first entry
			// in the table is normal alleles, and then the bad alleles in the same order as in the badAllele array
			if (vRaces != null)
				this.alleleFreqRates = new double[vRaces.size()][iNumOfAlleles];
			else
				this.alleleFreqRates = new double[1][iNumOfAlleles];

			for (int x = 0; x < alleleFreqRates.length; x++) {
				for (int y = 0; y < alleleFreqRates[0].length; y++)
					alleleFreqRates[x][y] = Double.valueOf(st.nextToken()).doubleValue();
			}

			// now that all the input variables are set, we can init all other locals
			init();

			System.out.println(this.toString());
		}
		catch(NoSuchElementException e) {
			throw new InvalidDiseaseException();
		}
		catch(NumberFormatException e) {
			throw new InvalidDiseaseException();
		}
	}


	private void init() {

		if (mutationRates == null)
			bHasMutationRate = false;
		else
			bHasMutationRate = true;

		if (expressionTable.length > 1)
			bAgeDependentExpression = true;
		else
			bAgeDependentExpression = false;

		if (vRaces == null)
			bRaceDependent = false;
		else
			bRaceDependent = true;

		// create the hashtable for the mutationIndex
		mutationAlleleIndexTable = new Hashtable(badAlleles.length);
		for(int x = 0; x < badAlleles.length; x++)
			mutationAlleleIndexTable.put(badAlleles[x], new Integer(x));

		// noMutationRate calculation
		noMutationRate = 1.0;
		if (bHasMutationRate) {
			for(int x=0; x < mutationRates.length; x++)
				noMutationRate = noMutationRate - mutationRates[x];
			//DEBUG
			System.out.println("No mutation rate: "+noMutationRate);
		}

		// create the genotypes and freqs
		int iNumOfAlleles = badAlleles.length + 1;
		genotypes = new String[iNumOfAlleles * (iNumOfAlleles + 1)/2];
		if (bRaceDependent)
			genotypeFreqs = new double[vRaces.size()][genotypes.length];
		else
			genotypeFreqs = new double[1][genotypes.length];
		
		int index=0;
		genotypes[index] = normalAllele + "-" + normalAllele;
		for(int x = 0; x < genotypeFreqs.length; x++)
			genotypeFreqs[x][index] = alleleFreqRates[x][index]*alleleFreqRates[x][index];
		// Ex. genotypeFreqs[race1][0] = alleleFreqRates[race1][0]^2
		index++; 
		for (int x =0; x < badAlleles.length; x++) {
			genotypes[index] = normalAllele + "-" + badAlleles[x];
			for(int y = 0; y < genotypeFreqs.length; y++)
				genotypeFreqs[y][index] = 2 * alleleFreqRates[y][0]*alleleFreqRates[y][x+1];
			index++;
		}
		for(int x=0; x < badAlleles.length; x++) {
			for(int y=x; y < badAlleles.length; y++) {
				genotypes[index] = badAlleles[x] + "-" + badAlleles[y];
				if (x == y) {
					for(int z = 0; z < genotypeFreqs.length; z++)
						genotypeFreqs[z][index] = alleleFreqRates[z][x+1]*alleleFreqRates[z][y+1];
				}
				else{
					for(int z = 0; z < genotypeFreqs.length; z++)
						genotypeFreqs[z][index] = 2 * alleleFreqRates[z][x+1]*alleleFreqRates[z][y+1];
				}
				index++;
			}
		}

		// create the inhertience table which has an entry for each combination of parent genotypes
		// size of inheritenceTable is taking two elements from genotypes with repetition
		// the array is set to all zeros by java
		inheritenceTable = new double[(genotypes.length+1)*(genotypes.length)/2][genotypes.length];
		int theIndexInheritenceTable = 0;
		for(int x = 0; x < genotypes.length; x++) {
			int iIndex = genotypes[x].indexOf('-');
			String alleleOne = genotypes[x].substring(0,iIndex);
			String alleleTwo = genotypes[x].substring(iIndex + 1);
			for (int y = x; y < genotypes.length; y++) {
				iIndex = genotypes[y].indexOf('-');
				String alleleThree = genotypes[y].substring(0, iIndex);
				String alleleFour =  genotypes[y].substring(iIndex + 1);
				setInheritenceTableEntry(inheritenceTable[theIndexInheritenceTable], alleleOne, alleleThree);
				setInheritenceTableEntry(inheritenceTable[theIndexInheritenceTable], alleleOne, alleleFour);
				setInheritenceTableEntry(inheritenceTable[theIndexInheritenceTable], alleleTwo, alleleThree);
				setInheritenceTableEntry(inheritenceTable[theIndexInheritenceTable], alleleTwo, alleleFour);
				theIndexInheritenceTable++;
			}
		}
		
	}

  public String getName(){
    return name;
  }

	private int getMutationIndex(String badAllele) {
	// requires:  mutationAlleleIndexTable has been initialized with all the badAlleles and their indexes by init()
		Integer iResult = (Integer)mutationAlleleIndexTable.get(badAllele);
		return iResult.intValue();
	}

	private boolean isNormalAllele(String allele) {
		if (allele.equals(normalAllele))
			return true;
		else return false;
	}

	// given the table entry and the two alleles, the genotypes that can result (taking mutation into account)
	// is added to the entry for the resulting genotypes
  private void setInheritenceTableEntry(double[] probEntry , String alleleOne, String alleleTwo) {
    System.out.println("Inheritence Table Entries for "+alleleOne+"-"+alleleTwo);  // DEBUG
    boolean normalOne = isNormalAllele(alleleOne);
    boolean normalTwo = isNormalAllele(alleleTwo);
    if (normalOne || normalTwo) {
      
      if (normalOne && normalTwo) {
	int index = 0;
	
	if (bHasMutationRate) {
	  probEntry[index] += .25 * noMutationRate * noMutationRate;  //ex. A-A
	  System.out.println("Prob Entry "+index+" value: "+probEntry[index]);  // DEBUG
	  for (int x = 0; x < badAlleles.length; x++){
	    index++;
	    probEntry[index] += .5 * noMutationRate * mutationRates[x];  //ex. A-b
	  // DEBUG
	  System.out.println("Prob Entry "+index+" "+probEntry[index]);
	  }
	  for(int x = 0; x < badAlleles.length; x++) {
	    for(int y = x; y < badAlleles.length; y++) {
	      if (x == y){
		index++;
		probEntry[index] += .25 * mutationRates[x] * mutationRates[x];  //ex. b-b
		System.out.println("Prob Entry "+index+" value: "+probEntry[index]); // DEBUG
	      }
	      else{
		index++;
		probEntry[index] += .5 * mutationRates[x] * mutationRates[y];  //ex. b-c
		System.out.println("Prob Entry "+index+" value: "+probEntry[index]); // DEBUG
	      }
	    }
	  }
	}
	else{  // No mutation rate, both alleles normal
	  probEntry[index] += .25;
	  System.out.println("Prob Entry "+index+" value: "+probEntry[index]); // DEBUG
	}
      }
      else {
				// only one is normal
	String theNormalAllele;
	String theMutationAllele;
	if (normalOne) {
	  theNormalAllele = alleleOne;
	  theMutationAllele = alleleTwo;
	}
	else {
	  theNormalAllele = alleleTwo;
	  theMutationAllele = alleleOne;
	}
	int indexOfMutationAllele = getMutationIndex(theMutationAllele);
	
	if (bHasMutationRate) {
	  probEntry[1+indexOfMutationAllele] += .25 * noMutationRate;  //ex. A-b
	  System.out.println("Prob Entry "+(1+indexOfMutationAllele)+" value: "+probEntry[1+indexOfMutationAllele]); // DEBUG
	  int indexIntoGenotypeArray = 0;
	  for (int x = 0;	x <= indexOfMutationAllele; x ++) {
	    indexIntoGenotypeArray += badAlleles.length - x + 1;  //Index for start of a row ex. a-a for a-a, a-b
	    probEntry[indexIntoGenotypeArray + indexOfMutationAllele - x] += .25 * mutationRates[x]; //Ex. a-b
	    System.out.println("Prob Entry "+(indexIntoGenotypeArray+indexOfMutationAllele-x)+" value: "+probEntry[indexIntoGenotypeArray+indexOfMutationAllele-x]); // DEBUG
	  }
	  for(int x=1; x < badAlleles.length - indexOfMutationAllele - 1; x++){
	    probEntry[indexIntoGenotypeArray + x] += 0.25 * mutationRates[x + indexOfMutationAllele];  //Ex. b-c
	    System.out.println("Prob Entry "+(indexIntoGenotypeArray+x)+" value: "+probEntry[indexIntoGenotypeArray+x]);  // DEBUG
	  }
	}
	else{  // no mutation rate
	  probEntry[1+indexOfMutationAllele] += .25; //Ex. A-b
	  System.out.println("Prob Entry "+(1+indexOfMutationAllele)+" value: "+probEntry[1+indexOfMutationAllele]); // DEBUG
	}
      }
    }
    else {  // no normal alleles, no mutation's can occur
      int indexAlleleOne = getMutationIndex(alleleOne);
      int indexAlleleTwo = getMutationIndex(alleleTwo);
      
      // make the first one the higher allele by swapping
      if (indexAlleleOne < indexAlleleTwo) {
	int temp = indexAlleleOne;
	indexAlleleOne = indexAlleleTwo;
	indexAlleleTwo = temp;
      }
      
      // only deal with indexes from here on
      int indexIntoGenotypeArray = 0;
      for(int i = 0; i <= indexAlleleTwo; i++){
	indexIntoGenotypeArray += badAlleles.length-i+1;
      }
      probEntry[indexIntoGenotypeArray + indexAlleleOne - indexAlleleTwo] += 0.25;
      System.out.println("Prob Entry "+(indexIntoGenotypeArray+indexAlleleOne-indexAlleleTwo)+" value: "+probEntry[indexIntoGenotypeArray+indexAlleleOne-indexAlleleTwo]);  // DEBUG
    }
  }
      
	public String[] getGenotypes(int pPersonSexType) {
		return genotypes;
	}

	public String[] getPhenotypes(int pPersonSexType) {
		String[] phenotypes = {"Unaffected", "Affected"};
		return phenotypes;
	}

	public void addPerson(BayesNetwork bn, Person indiv) {
		try {
		BayesNetworkNode genotypeNode;
		if (bRaceDependent) 
			genotypeNode = bn.createNode(genotypes, genotypeFreqs[indexForRace(indiv)]);
		else
			genotypeNode = bn.createNode(genotypes, genotypeFreqs[0]);

		System.out.println("created genotypeNode");
		double[] tempPhenoFreq = {1.0, 0};
		String[] phenotypes = {"Unaffected", "Affected"};
		BayesNetworkNode phenotypeNode = bn.createNode(phenotypes, tempPhenoFreq);
		bn.createArc(genotypeNode, phenotypeNode);

		// need to set up the expression of phenotype based on genotype
		double[] expressionTableEntry;
		if (bAgeDependentExpression)
			expressionTableEntry = expressionTable[getAgeIndex(indiv)];
		else
			expressionTableEntry = expressionTable[0];
		BayesNetworkNode[] parentNodes = {genotypeNode};
		double[] expressionProbs = new double[2];
		String[] parentValues = new String[1];
		for(int x = 0; x < genotypes.length; x++) {
			expressionProbs[1] = expressionTableEntry[x];
			expressionProbs[0] = 1-expressionProbs[1];
			parentValues[0] = genotypes[x];
			phenotypeNode.changeProbFunction(parentNodes, parentValues, expressionProbs);
		}

		indiv.setGenotypeNode(genotypeNode);
		indiv.setPhenotypeNode(phenotypeNode);
		}
		catch(ProbabilityException e) {
			System.err.println("Disease add person had a prob exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add person had a value exception");
		}
		catch(InvalidArcException e) {
			System.err.println("Disease add person had a arc exception");
		}
	}

	private int getAgeIndex(Person p) {
		return p.getAge() / 10;
	}

	public void removePerson(BayesNetwork bn, Person indiv) {
		// make sure no children or parents of gentype
		BayesNetworkNode genotypeNode = indiv.getGenotypeNode();
		Enumeration enumParents = genotypeNode.getParents();
		if (enumParents.hasMoreElements())
			return;  // don't do anything if has indiv has parents
		BayesNetworkNode phenotypeNode = indiv.getPhenotypeNode();
		Enumeration enumChildren = genotypeNode.getChildren();
		// must have at least one child, phenotype node
		enumChildren.nextElement();
		if (enumChildren.hasMoreElements())
			return;  // indiv has children

		bn.removeNode(phenotypeNode);
		bn.removeNode(genotypeNode);

		indiv.setPhenotypeNode(null);
		indiv.setGenotypeNode(null);
	}
	

	public void addChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// try making links
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();
		try {
			bn.createArc(parentNode1, childNode);
			bn.createArc(parentNode2, childNode);
		}
		catch (InvalidArcException e) {
			System.err.println("Disease.addChild had an arc exception");
			return;  // created a loop - time to bail
		}

		// need to reinit child genotypenode
		try {
			BayesNetworkNode[] bnns = {parentNode1, parentNode2};
			int theIndexInheritenceTable = 0;
			String[] parentValues = new String[2];
			for(int x = 0; x < genotypes.length; x++) {
				for (int y = x; y < genotypes.length; y++) {
					parentValues[0] = genotypes[x];
					parentValues[1] = genotypes[y];
					childNode.changeProbFunction(bnns, parentValues, inheritenceTable[theIndexInheritenceTable]);
					parentValues[0] = genotypes[y];
					parentValues[1] = genotypes[x];
					childNode.changeProbFunction(bnns, parentValues, inheritenceTable[theIndexInheritenceTable]);
					theIndexInheritenceTable++;
				}
			}
		}
		catch(ProbabilityException e) {
			System.err.println("Disease add child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add child had a value exception.");
		}
	}

	public void removeChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// remove the links if they exist
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();

		bn.removeArc(parentNode1, childNode);
		bn.removeArc(parentNode2, childNode);

		// init the genotype node again
		try {
		  if(bRaceDependent){
			childNode.changeProbFunction(null, null, genotypeFreqs[indexForRace(child)]);
		  }
		  else{
		    childNode.changeProbFunction(null, null, genotypeFreqs[0]);
		  }
		}
		catch(ProbabilityException e) {
			System.err.println("Disease remove child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease remove child had a value exception");
		}
	}

	public String toString() {
		String s = "NewAutosomalDisease\n";
		s += name + "\n";
		s += normalAllele + "\n";
		s += badAlleles.length + "\n";
		for (int x= 0; x < badAlleles.length; x++)
			s += badAlleles[x] + "\n";

		if (bHasMutationRate) {
			s += "Mutation Rates\n";
			for (int x = 0; x < mutationRates.length; x++)
				s += mutationRates[x] + "\n";
		}

		if (bAgeDependentExpression)
			s += "Age dependent\n";
		else
			s += "Not age dependent\n";
		for (int x= 0; x < expressionTable.length; x++) {
			for (int y = 0; y < expressionTable[0].length; y ++)
				s += expressionTable[x][y] + "\n";
		}
		
		
		if (bRaceDependent) {
			s += "Race dependent\n";
			s += vRaces.size() + "\n";
			for (int x=0; x < vRaces.size(); x++)
				s += vRaces.elementAt(x) + "\n";
		}

		s += "Freq rates\n";
		for (int x = 0; x < alleleFreqRates.length; x++) {
			for (int y = 0; y < alleleFreqRates[0].length; y++)
				s += alleleFreqRates[x][y] + "\n";
		}

		return s;
	}

	private int indexForRace(Person p) {
	// requires that person has a valid race 
		String sRace = p.getRace();
		for(int x = 0; x < vRaces.size(); x++)
			if (sRace.equals(vRaces.elementAt(x)))
			return x;

		return 0;  // this should never happen
	}

	public boolean hasMutationRate() {
		return bHasMutationRate;
	}

	public boolean isRaceDependent() {
		return bRaceDependent;
	}

	public String[] getRaces() {
		if (vRaces == null)
			return null;
		else {
			String[] races = new String[vRaces.size()];
			vRaces.copyInto(races);
			return races;
		}
	}

	public boolean isAgeDependent() {
		return bAgeDependentExpression;
	}

	public String getNormalAllele() {
		return normalAllele;
	}

	public String[] getDefectiveAlleles() {
		return badAlleles;
	}

	public boolean isSexLinked() {
		return false;
	}

	public double[] getMutationRates() {
		return mutationRates;
	}

	public double[][] getFrequencyRates() {
		return alleleFreqRates;
	}

	public double[][] getExpressionRates(int iSexType) {
		return expressionTable;
	}


}
