//package Pedigree;

//import Person.*;
//import BayesNetwork.*;
//import Disease.*;
//import GeneticScreen.*;
//import GenExceptions.*;

// this class maintains all the probabalistic information about the pedigree

import java.util.*;

public class Pedigree {

	Vector pedPeople;
	BayesNetwork bayesNet;
	Disease disease;
	boolean bUpdated;


// Constructors
	public Pedigree(Disease d) {
		pedPeople = new Vector();
		bayesNet = new GNUJavaBayesNet();
		disease = d;
		bUpdated = false;
	}


// methods
	public void addPerson(Person indiv) {
	// requires: indiv is not null, has an ID that is different from other people
	// effects: adds this person to pedigree and sets up his probabalistic information
		// need to make sure indiv is not already in the vector
		PedigreePerson pedperTemp = new PedigreePerson(indiv);
		if (!pedPeople.contains(pedperTemp)) {
			pedPeople.addElement(pedperTemp);
			disease.addPerson(bayesNet, indiv);
			clearAllInferences();
		}
	}

	
	public void removePerson(Person indiv) throws InvalidRelationshipException{
		// requires: indiv is not null
		// effects: remove this person from the pedigree if he is a member of the pedigree.
		//	  if you attempt to remove someone who is married or had children or has parents, 
		//    an InvalidRelationshipException occurs
		PedigreePerson pedper = lookupPedPerson(indiv);
		if (pedper == null)
			return;

		// only need to check for spouses since this must be true for children to exist
		Enumeration eSpouses = pedper.getSpouses();
		if (eSpouses.hasMoreElements())
			throw new InvalidRelationshipException();

		// also check for parents
		Enumeration eParents = pedper.getParents();
		if (eParents.hasMoreElements())
			throw new InvalidRelationshipException();

		pedPeople.removeElement(pedper);
		disease.removePerson(bayesNet, indiv);
		clearAllInferences();
	}

	// throws invalid relationship if parents can't be married or child creates loop or if child already has parents
	// if any of these people are not in pedigree they are added and the marriage is created, asuming
	// the whole operation is legal
	public void makeChild(Person child, Person parent1, Person parent2) throws InvalidRelationshipException {
		// first lets make sure parent1 and parent2 can be married (this may cause an exception), if so do it
		// then make sure child does not create a loop, is so undo new marriage (if it was new)
		// and throw the exception
		
		// are they already married?
		boolean alreadyMarried = false;
		PedigreePerson pedPerParent1 = lookupPedPerson(parent1);
		PedigreePerson pedPerParent2 = null;
		
		// see if they are married
		if (pedPerParent1 != null) {
			Enumeration ePedPer = pedPerParent1.getSpouses();
			while (ePedPer.hasMoreElements()) {
				PedigreePerson pedperSpouse = (PedigreePerson)ePedPer.nextElement();
				if (pedperSpouse.getPerson().equals(parent2)) {
					pedPerParent2 = pedperSpouse;
					alreadyMarried = true;
					break;
				}
			}
		}

		if (!alreadyMarried)
			pedPerParent2 = lookupPedPerson(parent2);

		// if they weren't already married, it may be because one or both parents don't exist in pedigree
		boolean parent1AlreadyExists = true;
		boolean parent2AlreadyExists = true;
		if (!alreadyMarried) {
			if (pedPerParent1 == null) {
				parent1AlreadyExists = false;
				pedPeople.addElement(pedPerParent1 = new PedigreePerson(parent1));
			}
			if (pedPerParent2 == null) {
				parent2AlreadyExists = false;
				pedPeople.addElement(pedPerParent2 = new PedigreePerson(parent2));
			}

			// now lets try adding the marriage
			try {
				// this can definately be done more efficiently
				makeMarriage(parent1, parent2);
			}
			catch (InvalidRelationshipException e) {
				// ok, marriage failed, now I have to remove all people I added
				if (!parent1AlreadyExists)
					pedPeople.removeElement(pedPerParent1);
				if (!parent2AlreadyExists)
					pedPeople.removeElement(pedPerParent2);
				throw e;
			}
		}



		// the marriage exists or now exists, so time to add the child
		PedigreePerson pedPerChild = lookupPedPerson(child);

		if (pedPerChild == null)  // if child did not exist, then it can definately become a child
			pedPeople.addElement(pedPerChild = new PedigreePerson(child));

		else {
			// this child exists, make sure he doesn't already have parents or will create a loop

			// make sure child doesn't already have parents
			Enumeration eParents = pedPerChild.getParents();
			if (eParents.hasMoreElements()) {
				// this relationship is invalid, so undo everything
				if (!alreadyMarried) {
					pedPerParent1.removeSpouse(pedPerParent2);
					pedPerParent2.removeSpouse(pedPerParent1);
				}
				if (!parent1AlreadyExists)
						pedPeople.removeElement(pedPerParent1);
				if (!parent2AlreadyExists)
						pedPeople.removeElement(pedPerParent2);

				throw new InvalidRelationshipException();
			}


			// make sure this is not a self loop
			if (pedPerParent1.equals(pedPerChild) || pedPerParent2.equals(pedPerChild)) {
				if (!alreadyMarried) {
						pedPerParent1.removeSpouse(pedPerParent2);
						pedPerParent2.removeSpouse(pedPerParent1);
					}
					if (!parent1AlreadyExists)
						pedPeople.removeElement(pedPerParent1);
					if (!parent2AlreadyExists)
						pedPeople.removeElement(pedPerParent2);

					throw new InvalidRelationshipException();
			}


			// make sure we will not create a loop
			Enumeration eDescendents = getDescendents(pedPerChild.getPerson());
			while(eDescendents.hasMoreElements()) {
				Person perDesc = (Person)eDescendents.nextElement();
				if (perDesc.equals(pedPerParent1.getPerson()) || perDesc.equals(pedPerParent2.getPerson())) {
					// we indeed have a loop, undo everything
					if (!alreadyMarried) {
						pedPerParent1.removeSpouse(pedPerParent2);
						pedPerParent2.removeSpouse(pedPerParent1);
					}
					if (!parent1AlreadyExists)
						pedPeople.removeElement(pedPerParent1);
					if (!parent2AlreadyExists)
						pedPeople.removeElement(pedPerParent2);

					throw new InvalidRelationshipException();
				}
			}
		}

		// now its ok to add the child
		pedPerChild.addParents(pedPerParent1, pedPerParent2);
		pedPerParent1.addChild(pedPerChild);
		pedPerParent2.addChild(pedPerChild);
		
		disease.addChild(bayesNet, child, parent1, parent2);
		clearAllInferences();
	}


	// makes sure a marriage is valid (sexes are opposite).  you can't marry a couple with unknown sex
	// adds the people if they don't exist, also does nothing if this marriage connection already exists
	public void makeMarriage(Person person1, Person person2) throws InvalidRelationshipException {
		if (((person1.getSex() == Person.MALE) && (person2.getSex() == Person.FEMALE)) ||
			((person1.getSex() == Person.FEMALE) && (person2.getSex() == Person.MALE))) {
			PedigreePerson pedper1 = lookupPedPerson(person1);
			if (pedper1 == null) {
				pedPeople.addElement(pedper1 = new PedigreePerson(person1));
				clearAllInferences();
			}
			PedigreePerson pedper2 = lookupPedPerson(person2);
			if (pedper2 == null) {
				pedPeople.addElement(pedper2 = new PedigreePerson(person2));
				clearAllInferences();
			}
			pedper1.addSpouse(pedper2);
			pedper2.addSpouse(pedper1);
		}
		else
			throw new InvalidRelationshipException();
	}

	// throws an exception is this couple has children.
	// does nothing if this couple is not married
	public void removeMarriage(Person person1, Person person2) throws InvalidRelationshipException {
		PedigreePerson pedper1 = lookupPedPerson(person1);
		PedigreePerson pedper2 = lookupPedPerson(person2);
		if ((pedper1 == null) || (pedper2 == null))
			return;

		if (!pedper1.isMarriedTo(pedper2))
			return;

		// make sure this couple doesn't have any childen
		Enumeration eP1Children = pedper1.getChildren();
		while (eP1Children.hasMoreElements()) {
			// for each kid, see if parent2 if a parent too
			PedigreePerson kid = (PedigreePerson)eP1Children.nextElement();
			Enumeration eParents = kid.getParents();
			// this kid better have parents
			PedigreePerson p1 = (PedigreePerson)eParents.nextElement();
			PedigreePerson p2 = (PedigreePerson)eParents.nextElement();
			if (p1.equals(pedper2) || p2.equals(pedper2))
				throw new InvalidRelationshipException();
		}

		// if we got here, these parents have no children together
		pedper1.removeSpouse(pedper2);
		pedper2.removeSpouse(pedper1);
	}

	// remove child does nothing if this person is not a child
	public void removeChild(Person child) {
		PedigreePerson pedperChild = lookupPedPerson(child);

		if (pedperChild == null)
			return;

		Enumeration eParents = pedperChild.getParents();
		if (eParents.hasMoreElements()) {
			PedigreePerson parent1 = (PedigreePerson)eParents.nextElement();
			parent1.removeChild(pedperChild);
			PedigreePerson parent2 = (PedigreePerson)eParents.nextElement();
			parent2.removeChild(pedperChild);
			pedperChild.removeParents();
			disease.removeChild(bayesNet, child, parent1.getPerson(), parent2.getPerson());
			clearAllInferences();
		}

	}

	public GeneticScreen createGeneticScreen(Person p, GeneticScreenDescriptor gsd) throws InvalidGeneticScreenDescriptorException{
		GeneticScreen gs = gsd.createGeneticScreen(this, p, bayesNet);
		p.addGeneticScreen(gs);
		clearAllInferences();
		return gs;
	}

	public void removeGeneticScreen(GeneticScreen gs) {
		gs.getGeneticScreenDescriptor().removeGeneticScreen(gs, bayesNet);
		gs.getPerson().removeGeneticScreen(gs);
		clearAllInferences();
	}

	public Enumeration getPeople() {
		// effects: Return the people in this pedigree
		Vector vPeople = new Vector();
		for(int x = 0; x < pedPeople.size(); x++) {
			PedigreePerson p = (PedigreePerson)pedPeople.elementAt(x);
			vPeople.addElement(p.getPerson());
		}
		return vPeople.elements();
	}


	public Enumeration getSpouses(Person indiv) {
		// requires: indiv is not null
		// effects:  Return the spouses of this indiv.  Returns an empty enumeration if indiv not in pedigree.
		Vector vSpouses = new Vector();

		PedigreePerson pedPer = lookupPedPerson(indiv);
		if (pedPer == null)
			return vSpouses.elements();  // empty enumeration

		Enumeration eSpouses = pedPer.getSpouses();
		while (eSpouses.hasMoreElements()) {
			PedigreePerson p1 = (PedigreePerson)eSpouses.nextElement();
			vSpouses.addElement(p1.getPerson());
		}

		return vSpouses.elements();		
	}

	public Enumeration getParents(Person indiv) {
		// requires: indiv is not null
		// effects:  Return the parents of this indiv.  Returns an empty enumeration if indiv not in pedigree.
		Vector vParents = new Vector();

		PedigreePerson pedPer = lookupPedPerson(indiv);
		if (pedPer == null)
			return vParents.elements();  // empty enumeration

		Enumeration eParents = pedPer.getParents();
		while (eParents.hasMoreElements()) {
			PedigreePerson p1 = (PedigreePerson)eParents.nextElement();
			vParents.addElement(p1.getPerson());
		}

		return vParents.elements();	
	}

	public Enumeration getChildren(Person indiv) {
		// requires: indiv is not null
		// effects:  Return the children of this indiv.  Returns an empty enumeration if indiv not in pedigree.
		Vector vChildren = new Vector();

		PedigreePerson pedPer = lookupPedPerson(indiv);
		if (pedPer == null)
			return vChildren.elements();  // empty enumeration

		Enumeration eChildren = pedPer.getChildren();
		while (eChildren.hasMoreElements()) {
			PedigreePerson p1 = (PedigreePerson)eChildren.nextElement();
			vChildren.addElement(p1.getPerson());
		}

		return vChildren.elements();
	}

	public Enumeration getChildren(Person person1, Person person2) {
		// requires: person1 and person2 is not null
		// effects:  Return the children of this couple.  Returns an empty enumeration if couple not in pedigree.
		PedigreePerson pedper1 = lookupPedPerson(person1);
		PedigreePerson pedper2 = lookupPedPerson(person2);

		Vector children = new Vector();

		if ((pedper1 == null) || (pedper2 == null))
			return children.elements();  // an empty enumeration

		if (!pedper1.isMarriedTo(pedper2))
			return children.elements();   // an empty enumeration

		Enumeration eP1Children = pedper1.getChildren();
		while (eP1Children.hasMoreElements()) {
			// for each kid, see if parent2 is a parent too
			PedigreePerson kid = (PedigreePerson)eP1Children.nextElement();
			Enumeration eParents = kid.getParents();
			// this kid better have parents
			PedigreePerson p1 = (PedigreePerson)eParents.nextElement();
			PedigreePerson p2 = (PedigreePerson)eParents.nextElement();
			if (p1.equals(pedper2) || p2.equals(pedper2))
				children.addElement(kid.getPerson());
		}

		return children.elements();
	}


	public Enumeration getDescendents(Person indiv) {
		// requires: indiv is not null
		// effects:  Return the descendents of this indiv.  Returns an empty enumeration if indiv not in pedigree.
		Vector vPeopleDesc = new Vector();
		PedigreePerson pedPerRoot = lookupPedPerson(indiv);
		if (pedPerRoot == null)
			return vPeopleDesc.elements();  // empty enumeration

		addDescendents(vPeopleDesc, pedPerRoot);
		
		return vPeopleDesc.elements();
	}


	private void addDescendents(Vector vPepDesc, PedigreePerson pedPerRoot) {
		// requires: inputs can't be null
		// effects: depth first addition of descendents are added to vector vPepDesc from the given root
		Enumeration enumChildren = pedPerRoot.getChildren();
		while(enumChildren.hasMoreElements()) {
			PedigreePerson pedPerChild = (PedigreePerson)enumChildren.nextElement();
			if (!vPepDesc.contains(pedPerChild.getPerson())) {
				vPepDesc.addElement(pedPerChild.getPerson());
				addDescendents(vPepDesc, pedPerChild);
			}
		}
	}

	public Enumeration getAncestors(Person indiv) {
		// requires: indiv is not null
		// effects:  Return the ancestors of this indiv.  Returns an empty enumeration if indiv not in pedigree.
		Vector vPeopleAncest = new Vector();
		PedigreePerson pedPerRoot = lookupPedPerson(indiv);
		if (pedPerRoot == null)
			return vPeopleAncest.elements();  // empty enumeration

		addAncestors(vPeopleAncest, pedPerRoot);
		
		return vPeopleAncest.elements();
	}

	private void addAncestors(Vector vPepAncest, PedigreePerson pedPerRoot) {
		// requires: inputs are not null
		// effects: depth first addition of ancestors are added to vector vPepAncest from the given root
		Enumeration enumParents = pedPerRoot.getParents();
		while(enumParents.hasMoreElements()) {
			PedigreePerson pedPerParent = (PedigreePerson)enumParents.nextElement();
			if (!vPepAncest.contains(pedPerParent.getPerson())) {
				vPepAncest.addElement(pedPerParent.getPerson());
				addAncestors(vPepAncest, pedPerParent);
			}
		}
	}

	public void updatePersonInferences() throws ProbabilityException{
		if (!bUpdated) {  // only do this if inferences are out of date
			for (int x = 0; x < pedPeople.size(); x++) {
				Person indiv = ((PedigreePerson) pedPeople.elementAt(x)).getPerson();
				System.out.println(indiv);
				System.out.println("Genotype Node");
				System.out.println(indiv.getGenotypeNode());
				indiv.setGenotypeInferences(bayesNet.inferenceOnNode(indiv.getGenotypeNode()));
				System.out.println("Updated Genotype Inferences");
				System.out.println("Phenotype Node");
				System.out.println(indiv.getPhenotypeNode());
				indiv.setPhenotypeInferences(bayesNet.inferenceOnNode(indiv.getPhenotypeNode()));
				System.out.println("Updated Phenotype Inferences");
			}
			bUpdated = true;
		}
	}

	public void clearAllInferences() {
		// add a variable so that I don't do this if there are no inferences ...
		if (bUpdated) {
			for (int x = 0; x < pedPeople.size(); x++) {
				Person indiv = ((PedigreePerson) pedPeople.elementAt(x)).getPerson();
				indiv.setGenotypeInferences(null);
				indiv.setPhenotypeInferences(null);
			}
		}
		bUpdated = false;
	}

	private PedigreePerson lookupPedPerson(Person p) {
		// requires: p is not null
		// effects: get the pedPerson that refers to this person p.  return null if p is not in pedigree.
		int index = pedPeople.indexOf(new PedigreePerson(p));
		if (index == -1)
			return null;
		else
			return (PedigreePerson)pedPeople.elementAt(index);
	}

}


class PedigreePerson {
	private Person personP;
	private Vector vpedperSpouses;
	private Vector vpedperChildren;
	private Vector vpedperParents;

	// Constructor
	public PedigreePerson(Person p) {
		personP = p;
		vpedperSpouses = new Vector();
		vpedperChildren = new Vector();
		vpedperParents = new Vector();
	}

	// methods
	public Person getPerson() {
		return personP;
	}

	public boolean equals(Object o) {
		try {
			PedigreePerson p = (PedigreePerson)o;
			return p.personP.equals(personP);
		}
		catch(ClassCastException e) {
			return false;
		}
	}

	// if this spouse is already a spouse, nothing is done
	public void addSpouse(PedigreePerson p) {
		if (!vpedperSpouses.contains(p))
			vpedperSpouses.addElement(p);
	}

	public void removeSpouse(PedigreePerson p) {
		vpedperSpouses.removeElement(p);
	}

	public Enumeration getSpouses() {
		return vpedperSpouses.elements();
	}

	// if this child is already a child, nothing is done
	public void addChild(PedigreePerson p) {
		if (!vpedperChildren.contains(p))
			vpedperChildren.addElement(p);
	}

	public void removeChild(PedigreePerson p) {
		vpedperChildren.removeElement(p);
	}

	public Enumeration getChildren() {
		return vpedperChildren.elements();
	}

	// a person can only have two parents at most, so this throws an exception if you 
	// try to have more than 2
	public void addParents(PedigreePerson p1, PedigreePerson p2) throws InvalidRelationshipException {
		if (vpedperParents.size() >= 2)
			throw new InvalidRelationshipException();
		vpedperParents.addElement(p1);
		vpedperParents.addElement(p2);
	}

	public void removeParents() {
		vpedperParents.removeAllElements();
	}

	// a person could have 0 or 2 parents depending on our knowldege of the pedigree
	public Enumeration getParents() {
		return vpedperParents.elements();
	}


	// checks for efficiency
	public boolean isChildOf(PedigreePerson parent) {
		return vpedperParents.contains(parent);
	}

	public boolean isParentOf(PedigreePerson child) {
		return vpedperChildren.contains(child);
	}

	public boolean isMarriedTo(PedigreePerson spouse) {
		return vpedperSpouses.contains(spouse);
	}


	public String toString() {
		return personP.toString() + ": Spouses: " + vpedperSpouses.toString() + " Children: " + 
			vpedperChildren.toString() + " Parents: " + vpedperParents.toString();
	}

}
