//package Disease;

//import Person.*;
//import BayesNetwork.*;
//import GenExceptions.*;

import java.util.*;

public class XlinkedDisease extends Disease {

private String	 name;
private String[] alleles;
private String[] mgenotypes;
private String[] fgenotypes;
private String[] unknownSexGenotypes;
private String[] phenotypes;
private double[] gFDefaultProbs;
private double[] gMDefaultProbs;
private double[] gUDefaultProbs;

private boolean bDomRec;

//constructor
	public XlinkedDisease(String name, String[] Alleles, String[] Phenotypes, double[] fprobs, double[] mprobs) 
		throws InvalidDiseaseException {
		// requires:  none of the inputs can be null
		// effects:  Creates an autosomal disease with the two alleles in Alleles, which
		//		have the 3 phenotypes in Phenotypes where Phenotypes[0] is the phenoptye for
		//      the genotype Allele1-Allele1, Phenotype[1] is for Allele1-Allele2, and Phenotype[2]
		//		is for Allele2-Allele2.  fprobs are the probabilities for of each phenotypes (and its
		//		genotypes since 100% penentrance) for women.  mprobs is probs for male phenotypes.
		// exceptions: An invaliddiseaseexception is thrown if alleles does not contain 2 different alleles,
		//		if phenotypes does not contain 3 items, if Phenotype[0] = phenotype[2], 
		//		if fprobs does not contain 3 double that sum close to 1, or if mprobs does not contain 2
		//      doubles that sum close to 1

		initXlinkedDisease(name, Alleles, Phenotypes, fprobs, mprobs);
	}

	private void initXlinkedDisease(String name, String[] Alleles, String[] Phenotypes, double[] fprobs, double[] mprobs) 
		throws InvalidDiseaseException {

		this.name = name;

		// need to put dominant gene first and its phenotype first
		if ((fprobs.length != 3) || (mprobs.length != 2) || (Alleles.length != 2) || (Alleles[0].equals(Alleles[1])))
			throw new InvalidDiseaseException("arguments incorrect");

		if ((Phenotypes.length != 3) || (Phenotypes[0].equals(Phenotypes[2])))
			throw new InvalidDiseaseException("arguments for phenotypes do not meet requirements");
		if (Phenotypes[0].equals(Phenotypes[1])) {
			// this is domrec- first allele is dominant
			bDomRec = true;
			this.alleles = (String[])Alleles.clone();
			this.phenotypes = new String[2];
			this.phenotypes[0] = Phenotypes[0];
			this.phenotypes[1] = Phenotypes[2];
			this.gFDefaultProbs = (double[])fprobs.clone();
			this.gMDefaultProbs = (double[])mprobs.clone();
		}
		else if (Phenotypes[1].equals(Phenotypes[2])) {  // Second allele given is dominant
			bDomRec = true;
			this.alleles = new String[2];
			this.alleles[0] = Alleles[1];
			this.alleles[1] = Alleles[0];
			this.phenotypes = new String[2];
			this.phenotypes[0] = Phenotypes[2];
			this.phenotypes[1] = Phenotypes[0];
			this.gFDefaultProbs = new double[3];
			this.gFDefaultProbs[0] = fprobs[2];
			this.gFDefaultProbs[1] = fprobs[1];
			this.gFDefaultProbs[2] = fprobs[0];  // 3 for females
			this.gMDefaultProbs = new double[2];
			this.gMDefaultProbs[0] = mprobs[1];
			this.gMDefaultProbs[1] = mprobs[0];  // 2 for males
		}
		else {
			// this is a codom disease
			bDomRec = false;
			this.alleles = (String[]) Alleles.clone();
			this.phenotypes = (String[]) Phenotypes.clone();
			this.gFDefaultProbs = (double[]) fprobs.clone();
			this.gMDefaultProbs = (double[]) mprobs.clone();
		}

		fgenotypes = new String[3];
		fgenotypes[0] = Alleles[0] + "-" + Alleles[0];
		fgenotypes[1] = Alleles[0] + "-" + Alleles[1];
		fgenotypes[2] = Alleles[1] + "-" + Alleles[1];

		mgenotypes = new String[2];  // Only 1 x
		mgenotypes[0] = Alleles[0];
		mgenotypes[1] = Alleles[1];

		unknownSexGenotypes = new String[mgenotypes.length + fgenotypes.length];
		gUDefaultProbs = new double[unknownSexGenotypes.length];
		int x;
		for(x=0; x < fgenotypes.length; x++) {  // fgenotypes.length == 3
			unknownSexGenotypes[x] = fgenotypes[x];
			gUDefaultProbs[x] = gFDefaultProbs[x] * .5;  // .5 == probability of being female
		}
		for(int y=x; y < mgenotypes.length + fgenotypes.length; y++) {
			unknownSexGenotypes[y] = mgenotypes[y-x];
			gUDefaultProbs[y] = gMDefaultProbs[y-x] * .5;
		}


		// make sure probs are valid 
		if (!super.sumToOne(fprobs) || !super.sumToOne(mprobs))
			throw new InvalidDiseaseException();

		sA1A1_A1 = new String[2];  // Format: FemaleAlleles__MaleAllele
		sA1A1_A1[0] = fgenotypes[0];
		sA1A1_A1[1] = mgenotypes[0];
		sA1A1_A2 = new String[2];
		sA1A1_A2[0] = fgenotypes[0];
		sA1A1_A2[1] = mgenotypes[1];
		sA1A2_A1 = new String[2];
		sA1A2_A1[0] = fgenotypes[1];
		sA1A2_A1[1] = mgenotypes[0];
		sA1A2_A2 = new String[2];
		sA1A2_A2[0] = fgenotypes[1];
		sA1A2_A2[1] = mgenotypes[1];
		sA2A2_A1 = new String[2];
		sA2A2_A1[0] = fgenotypes[2];
		sA2A2_A1[1] = mgenotypes[0];
		sA2A2_A2 = new String[2];
		sA2A2_A2[0] = fgenotypes[2];
		sA2A2_A2[1] = mgenotypes[1];
	}

	public XlinkedDisease(String sDisease) throws InvalidDiseaseException{
		// requires:  sDisease is not null
	  // sDisease must be in the format "Xlinked Disease" <name> <allele1> <allele2> <phenotype1> <phenotype2> <phenotype3> <fprob1> <fprob2> <fprob3> <mprob1> <mmmprob2>
		// effects: an xlinked disease based on the specification in the string
		//			An invalidDiseaseException is thrown if the string contains and invalid xlinkedDisease
		StringTokenizer st = new StringTokenizer(sDisease);
		try {
			String sNext = st.nextToken();
			if (!sNext.equals("XlinkedDisease"))
				throw new InvalidDiseaseException();

			String name = st.nextToken();
			String[] sAlleles = new String[2];
			sAlleles[0] = st.nextToken();
			sAlleles[1] = st.nextToken();
			String[] sPhenotypes = new String[3];
			sPhenotypes[0] = st.nextToken();
			sPhenotypes[1] = st.nextToken();
			sPhenotypes[2] = st.nextToken();

			double[] fprobs = new double[3];
			fprobs[0] = Double.valueOf(st.nextToken()).doubleValue();
			fprobs[1] = Double.valueOf(st.nextToken()).doubleValue();
			fprobs[2] = Double.valueOf(st.nextToken()).doubleValue();
			double[] mprobs = new double[2];
			mprobs[0] = Double.valueOf(st.nextToken()).doubleValue();
			mprobs[1] = Double.valueOf(st.nextToken()).doubleValue();

			initXlinkedDisease(name, sAlleles, sPhenotypes, fprobs, mprobs);
		}
		catch(NoSuchElementException e) {
			throw new InvalidDiseaseException();
		}
		catch(NumberFormatException e) {
			throw new InvalidDiseaseException();
		}
	}


// methods
	public String getName() {
		return name;
	}

	public String[] getGenotypes(int sexType) {
		if (sexType == Person.MALE)
			return (String[])mgenotypes.clone();
		else if (sexType == Person.FEMALE)
			return (String[])fgenotypes.clone();
		else 
			return (String[])unknownSexGenotypes.clone();
	}

	public String[] getPhenotypes(int sexType) {
		if (bDomRec)
			return (String[])phenotypes.clone();
		else {
			if (sexType == Person.MALE) {
				// can't have hybrid genotype's phenotype
				String[] mPhen = new String[2];
				mPhen[0] = phenotypes[0];
				mPhen[1] = phenotypes[2];
				return mPhen;
			}
			else  //female and unknown sex
				return (String[])phenotypes.clone();
		}
	}

	public String getDefaultPhenotype(int sexType) {
		return phenotypes[0];
	}

	//******

  private final double[] gFA1A1_A1probs = {1, 0, 0};  // Format: (A1A1, A1A2, A2A2)
	private final double[] gFA1A1_A2probs = {0, 1, 0};
	private final double[] gFA1A2_A1probs = {.5, .5, 0};
	private final double[] gFA1A2_A2probs = {0, .5, .5};
	private final double[] gFA2A2_A1probs = {0, 1, 0};
	private final double[] gFA2A2_A2probs = {0, 0, 1};

  private final double[] gMA1A1_A1probs = {1, 0};// Format: A1 A2
	private final double[] gMA1A1_A2probs = {1, 0};
	private final double[] gMA1A2_A1probs = {.5, .5};
	private final double[] gMA1A2_A2probs = {.5, .5};
	private final double[] gMA2A2_A1probs = {0, 1};
	private final double[] gMA2A2_A2probs = {0, 1};

	private final double[] gUA1A1_A1probs = {.5, 0, 0, .5, 0};
	private final double[] gUA1A1_A2probs = {0, .5, 0, .5, 0};
	private final double[] gUA1A2_A1probs = {.25, .25, 0,.25, .25};
	private final double[] gUA1A2_A2probs = {0, .25, .25, .25, .25};
	private final double[] gUA2A2_A1probs = {0, .5, 0, 0, .5};
	private final double[] gUA2A2_A2probs = {0, 0, .5, 0, .5};


	private String[] sA1A1_A1;
	private String[] sA1A1_A2;				
	private String[] sA1A2_A1;		
	private String[] sA1A2_A2;				
	private String[] sA2A2_A1;		
	private String[] sA2A2_A2;

	public void addPerson(BayesNetwork bn, Person indiv) {
		// create the two nodes
		try {
			BayesNetworkNode phenotypeNode;

			if (indiv.getSex() == Person.MALE) {
				BayesNetworkNode genotypeNode = bn.createNode(getGenotypes(indiv.getSex()), gMDefaultProbs);
				// this is temp
				phenotypeNode= bn.createNode(getPhenotypes(indiv.getSex()), gMA1A1_A1probs);  // indiv == MALE
				indiv.setGenotypeNode(genotypeNode);
				indiv.setPhenotypeNode(phenotypeNode);
				bn.createArc(genotypeNode, phenotypeNode);
				BayesNetworkNode[] bnns = {genotypeNode};
				String[] sphA1 = {mgenotypes[0]};
				String[] sphA2 = {mgenotypes[1]};
				double[] tempPhenProbs1 = {1, 0};
				double[] tempPhenProbs2 = {0, 1};
				phenotypeNode.changeProbFunction(bnns, sphA1, tempPhenProbs1);  // bnns == genotypeNode
				phenotypeNode.changeProbFunction(bnns, sphA2, tempPhenProbs2);
			}
			else if (indiv.getSex() == Person.FEMALE) {
				BayesNetworkNode genotypeNode = bn.createNode(getGenotypes(indiv.getSex()), gFDefaultProbs);
				if (bDomRec) {
					double[] tempPhenProbs = {1, 0};
					phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
				}
				else {  // Codominant
					double[] tempPhenProbs = {1, 0, 0};
					phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
				}
				indiv.setGenotypeNode(genotypeNode);
				indiv.setPhenotypeNode(phenotypeNode);
				bn.createArc(genotypeNode, phenotypeNode);
				BayesNetworkNode[] bnns = {genotypeNode};
				String[] sphA1A1 = {fgenotypes[0]};
				String[] sphA1A2 = {fgenotypes[1]};
				String[] sphA2A2 = {fgenotypes[2]};
				if (bDomRec) {
					double[] phA1A1 = {1, 0};
					double[] phA1A2 = {1, 0};
					double[] phA2A2 = {0, 1};
					phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
					phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
				}
				else  {  // Codominant
					double[] phA1A1 = {1, 0, 0};
					double[] phA1A2 = {0, 1, 0};
					double[] phA2A2 = {0, 0, 1};
					phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
					phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
				}
			}
			else {
				// this is a unknown sex person
				BayesNetworkNode genotypeNode = bn.createNode(getGenotypes(indiv.getSex()), gUDefaultProbs);

				// this is temp
				if (bDomRec) {
					double[] tempPhenProbs = {1, 0};
					phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
				}
				else {
					double[] tempPhenProbs = {1, 0, 0};
					phenotypeNode = bn.createNode(getPhenotypes(indiv.getSex()), tempPhenProbs); 
				}

				indiv.setGenotypeNode(genotypeNode);
				indiv.setPhenotypeNode(phenotypeNode);
				bn.createArc(genotypeNode, phenotypeNode);
				BayesNetworkNode[] bnns = {genotypeNode};

				String[] sphA1A1 = {fgenotypes[0]};
				String[] sphA1A2 = {fgenotypes[1]};
				String[] sphA2A2 = {fgenotypes[2]};
				String[] sphA1 = {mgenotypes[0]};
				String[] sphA2 = {mgenotypes[1]};
				if (bDomRec) {
					double[] phA1A1 = {1, 0};
					double[] phA1A2 = {1, 0};
					double[] phA2A2 = {0, 1};
					phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
					phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
					phenotypeNode.changeProbFunction(bnns, sphA1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA2, phA2A2);
				}
				else  {  // Codominant
				  double[] phA1A1 = {1, 0, 0};  // phenotypes
					double[] phA1A2 = {0, 1, 0};
					double[] phA2A2 = {0, 0, 1};
					phenotypeNode.changeProbFunction(bnns, sphA1A1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA1A2, phA1A2);
					phenotypeNode.changeProbFunction(bnns, sphA2A2, phA2A2);
					phenotypeNode.changeProbFunction(bnns, sphA1, phA1A1);
					phenotypeNode.changeProbFunction(bnns, sphA2, phA2A2);
				}

								
			}
			phenotypeNode.makeObservation(getDefaultPhenotype(indiv.getSex()));
		}
		catch(ProbabilityException e) {
			System.err.println("Disease add person had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add person had a value exception");
		}
		catch(InvalidArcException e) {
			System.err.println("Disease add person had an exception");
		}
	}

	public void removePerson(BayesNetwork bn, Person indiv) { // only work if person is not connected
		// make sure no children or parents of genotype
		BayesNetworkNode genotypeNode = indiv.getGenotypeNode();
		Enumeration enumParents = genotypeNode.getParents();
		if (enumParents.hasMoreElements())
			return;  // don't do anything if has indiv has parents
		BayesNetworkNode phenotypeNode = indiv.getPhenotypeNode();
		Enumeration enumChildren = genotypeNode.getChildren();
		// must have at least one child, phenotype node
		enumChildren.nextElement();
		if (enumChildren.hasMoreElements())
			return;  // indiv has children

		bn.removeNode(phenotypeNode);
		bn.removeNode(genotypeNode);
	}

	public void addChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// try making links
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();
		try {
			bn.createArc(parentNode1, childNode);
			bn.createArc(parentNode2, childNode);
		}
		catch (InvalidArcException e) {
			System.err.println("XlinkDisease had arc exception.");
			return;  // created a loop - time to bail
		}

		// need to reinit child genotypenode
		try {
			BayesNetworkNode[] bnns = new BayesNetworkNode[2];
			if (parent1.getSex() == Person.MALE) {  // so that female parent always listed first
				bnns[0] = parentNode2;
				bnns[1] = parentNode1;
			}
			else {
				bnns[0] = parentNode1;
				bnns[1] = parentNode2;
			}

			if (child.getSex() == Person.MALE) {
				childNode.changeProbFunction(bnns, sA1A1_A1, gMA1A1_A1probs);
				childNode.changeProbFunction(bnns, sA1A1_A2, gMA1A1_A2probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gMA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gMA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A1, gMA2A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A2, gMA2A2_A2probs);
			}
			else if (child.getSex() == Person.FEMALE) {
				childNode.changeProbFunction(bnns, sA1A1_A1, gFA1A1_A1probs);
				childNode.changeProbFunction(bnns, sA1A1_A2, gFA1A1_A2probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gFA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gFA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A1, gFA2A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A2, gFA2A2_A2probs);
			}
			else { // unknown sex child
				childNode.changeProbFunction(bnns, sA1A1_A1, gUA1A1_A1probs);
				childNode.changeProbFunction(bnns, sA1A1_A2, gUA1A1_A2probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gUA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA1A2_A1, gUA1A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A1, gUA2A2_A1probs);
				childNode.changeProbFunction(bnns, sA2A2_A2, gUA2A2_A2probs);
			}

		}
		catch(ProbabilityException e) {
			System.err.println("Disease add child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease add child had a value exception.");
		}
	}

	public void removeChild(BayesNetwork bn, Person child, Person parent1, Person parent2) {
		// remove the links if they exist
		BayesNetworkNode childNode = child.getGenotypeNode();
		BayesNetworkNode parentNode1 = parent1.getGenotypeNode();
		BayesNetworkNode parentNode2 = parent2.getGenotypeNode();

		bn.removeArc(parentNode1, childNode);
		bn.removeArc(parentNode2, childNode);

		// init the genotype node again
		try {
			if (child.getSex() == Person.MALE)
				childNode.changeProbFunction(null, null, gMDefaultProbs);
			else if (child.getSex() == Person.FEMALE)
				childNode.changeProbFunction(null, null, gFDefaultProbs);
			else
				childNode.changeProbFunction(null, null, gUDefaultProbs);

		}
		catch(ProbabilityException e) {
			System.err.println("Disease remove child had an exception");
		}
		catch(InvalidValueException e) {
			System.err.println("Disease remove child had a value exception");
		}
	}

  /** effects: creates a string wich can be put into the Xlinked Disease function to create an identical disease
   **/
	public String toString() {
		String s = "XlinkedDisease\n";
		s += name + "\n";
		s += alleles[0] + "\n";
		s += alleles[1] + "\n";
		s += phenotypes[0] + "\n";
		if (bDomRec) {
			s += phenotypes[0] + "\n";
			s += phenotypes[1] + "\n";
		}
		else {
			s += phenotypes[1] + "\n";
			s += phenotypes[2] + "\n";
		}
		for (int x = 0; x < gFDefaultProbs.length; x++)
			s += gFDefaultProbs[x] + "\n";
		for (int x = 0; x < gMDefaultProbs.length; x++)
			s += gMDefaultProbs[x] + "\n";

		return s;
	}

	// **** methods specific to an xlinked disease

	public String[] getAlleles() {
		return (String[])alleles.clone();
	}

	public String[] getPhenotypes() {
		return (String[])phenotypes.clone();
	}

	public double[] getMaleGenoProbs() {
		return (double[])gMDefaultProbs.clone();
	}

	public double[] getFemaleGenoProbs() {
		return (double[])gFDefaultProbs.clone();
	}

  public double[][] getExpressionRates(int sexType){
    //requires: 100% penetrance
    //effects: returns the expression rates from the possible genotypes
    double[][] expressionRates = new double[1][3];
    if(sexType == Person.MALE){
      expressionRates[0][0] = 1;
      expressionRates[0][1] = 1;
      expressionRates[0][2] = 1;
      return expressionRates;
    }
    else if(sexType == Person.FEMALE){
      expressionRates[0][0] = 1;
      expressionRates[0][1] = 1;
      expressionRates[0][2] = 1;
      return expressionRates;
    }
    else {
      expressionRates[0][0] = 1;
      expressionRates[0][1] = 1;
      expressionRates[0][2] = 1;
      return expressionRates;
    }
  }

  public boolean isAgeDependent(){
    //effects: returns false for now.
    return false;
  }

  public double[][] getFrequencyRates(){
    //effects: returns 0.1 for now.
    double[][] frequencyRates = new double[1][2];
    frequencyRates[0][0] = 0.1;
    frequencyRates[0][1] = 0.1;
    return frequencyRates;
  }

  public boolean isRaceDependent(){
    //effects: returns false for now.
    return false;
  }

  public String[] getRaces(){
    //effects: returns null for now.
    return null;
  }

  public String getNormalAllele(){
    //effects: returns dominant allele for now
    return alleles[0];
  }

  public String[] getDefectiveAlleles(){
    //effects: returns all alleles for now
    return (String[])alleles.clone();
  }

  public boolean isSexLinked(){
    //effects: returns false for now
    return false;
  }

  public boolean hasMutationRate(){
    //effects: returns false for now
    return false;
  }

  public double[] getMutationRates(){
    //effects: returns 0 for 1 defective allele for now
    double[] mutationRates = new double[1];
    mutationRates[0] = 0;
    return mutationRates;
  }
  
}
